from dao_treasury import ENVIRONMENT_VARIABLES as ENVS
from dao_treasury._nicknames import setup_address_nicknames_in_db
from dao_treasury._wallet import TreasuryWallet
from dao_treasury.db import TreasuryTx
from dao_treasury.sorting import (
    CostOfRevenueSortRule,
    ExpenseSortRule,
    IgnoreSortRule,
    OtherExpenseSortRule,
    OtherIncomeSortRule,
    RevenueSortRule,
    SortRuleFactory,
    cost_of_revenue,
    expense,
    ignore,
    other_expense,
    other_income,
    revenue,
)
from dao_treasury.treasury import Treasury


setup_address_nicknames_in_db()


if ENVS.SQL_DEBUG:
    import pony.orm

    pony.orm.sql_debug(True)

__all__ = [
    "Treasury",
    "TreasuryWallet",
    "CostOfRevenueSortRule",
    "ExpenseSortRule",
    "IgnoreSortRule",
    "OtherExpenseSortRule",
    "OtherIncomeSortRule",
    "RevenueSortRule",
    "cost_of_revenue",
    "expense",
    "ignore",
    "other_expense",
    "other_income",
    "revenue",
    "TreasuryTx",
    "SortRuleFactory",
]



from typing import Any, Final, Generic, Optional, Union, final, overload

from dao_treasury.constants import CHAINID
from dao_treasury.sorting.rule import (
    CostOfRevenueSortRule,
    ExpenseSortRule,
    IgnoreSortRule,
    OtherExpenseSortRule,
    OtherIncomeSortRule,
    RevenueSortRule,
    TRule,
)
from dao_treasury.types import Networks, SortFunction, TxGroupName


def revenue(
    txgroup: TxGroupName, networks: Networks = CHAINID
) -> "SortRuleFactory[RevenueSortRule]":
    """Create a factory to register revenue sort rules for a given transaction group.

    Args:
        txgroup: Base name of the transaction group to categorize as revenue.
        networks: Network ID or iterable of network IDs on which this rule applies.

    See Also:
        :func:`cost_of_revenue`
        :class:`SortRuleFactory`

    Examples:
        >>> from dao_treasury.sorting.factory import revenue
        >>> @revenue("Token Sales")
        ... def match_sales(tx):
        ...     return tx.amount > 0 and tx.to_address is not None
    """
    return SortRuleFactory(txgroup, networks, RevenueSortRule)


def cost_of_revenue(
    txgroup: TxGroupName, networks: Networks = CHAINID
) -> "SortRuleFactory[CostOfRevenueSortRule]":
    """Create a factory to register cost‐of‐revenue sort rules for a given transaction group.

    Args:
        txgroup: Base name of the transaction group to categorize as cost of revenue.
        networks: Network ID or iterable of network IDs on which this rule applies.

    See Also:
        :func:`revenue`
        :class:`SortRuleFactory`

    Examples:
        >>> from dao_treasury.sorting.factory import cost_of_revenue
        >>> @cost_of_revenue("Manufacturing")
        ... def match_manufacturing(tx):
        ...     return tx.from_address is not None and tx.amount_usd > 1000
    """
    return SortRuleFactory(txgroup, networks, CostOfRevenueSortRule)


def expense(
    txgroup: TxGroupName, networks: Networks = CHAINID
) -> "SortRuleFactory[ExpenseSortRule]":
    """Create a factory to register expense sort rules for a given transaction group.

    Args:
        txgroup: Base name of the transaction group to categorize as expense.
        networks: Network ID or iterable of network IDs on which this rule applies.

    See Also:
        :func:`other_expense`
        :class:`SortRuleFactory`

    Examples:
        >>> from dao_treasury.sorting.factory import expense
        >>> @expense("Office Supplies")
        ... def match_supplies(tx):
        ...     return tx.symbol == "USD" and tx.amount < 500
    """
    return SortRuleFactory(txgroup, networks, ExpenseSortRule)


def other_income(
    txgroup: TxGroupName, networks: Networks = CHAINID
) -> "SortRuleFactory[OtherIncomeSortRule]":
    """Create a factory to register other‐income sort rules for a given transaction group.

    Args:
        txgroup: Base name of the transaction group to categorize as other income.
        networks: Network ID or iterable of network IDs on which this rule applies.

    See Also:
        :func:`revenue`
        :class:`SortRuleFactory`

    Examples:
        >>> from dao_treasury.sorting.factory import other_income
        >>> @other_income("Interest")
        ... def match_interest(tx):
        ...     return tx.token_address == SOME_TOKEN and tx.amount > 0
    """
    return SortRuleFactory(txgroup, networks, OtherIncomeSortRule)


def other_expense(
    txgroup: TxGroupName, networks: Networks = CHAINID
) -> "SortRuleFactory[OtherExpenseSortRule]":
    """Create a factory to register other‐expense sort rules for a given transaction group.

    Args:
        txgroup: Base name of the transaction group to categorize as other expense.
        networks: Network ID or iterable of network IDs on which this rule applies.

    See Also:
        :func:`expense`
        :class:`SortRuleFactory`

    Examples:
        >>> from dao_treasury.sorting.factory import other_expense
        >>> @other_expense("Misc Fees")
        ... def match_misc(tx):
        ...     return tx.amount_usd < 0 and tx.symbol == "ETH"
    """
    return SortRuleFactory(txgroup, networks, OtherExpenseSortRule)


def ignore(
    txgroup: TxGroupName, networks: Networks = CHAINID
) -> "SortRuleFactory[IgnoreSortRule]":
    """Create a factory to register ignore sort rules for a given transaction group.

    Args:
        txgroup: Base name of the transaction group to categorize as ignored.
        networks: Network ID or iterable of network IDs on which this rule applies.

    See Also:
        :class:`SortRuleFactory`

    Examples:
        >>> from dao_treasury.sorting.factory import ignore
        >>> @ignore("Dust")
        ... def match_dust(tx):
        ...     return abs(tx.value_usd) < 0.01
    """
    return SortRuleFactory(txgroup, networks, IgnoreSortRule)


@final
class SortRuleFactory(Generic[TRule]):
    """Builder for creating sort rule instances for a specific transaction group and network(s).

    This factory supports two patterns:

    1. Decorating a function to register a dynamic matching rule.
    2. Calling :meth:`match` to supply static match attributes.

    Use the convenience functions like :func:`revenue`, :func:`expense`, etc.,
    to obtain an instance of this factory preconfigured with the appropriate rule type.

    Examples:
        >>> from dao_treasury.sorting.factory import revenue
        >>> @revenue("Sales", networks=[1, 3])
        ... def match_large_sales(tx):
        ...     return tx.value_usd > 1000
    """

    def __init__(
        self,
        txgroup: TxGroupName,
        networks: Networks,
        rule_type: TRule,
    ) -> None:
        """Initialize the sort rule factory.

        Args:
            txgroup: Base name of the transaction group.
            networks: Single network ID or iterable of network IDs where the rule applies.
            rule_type: Sort rule class (e.g., RevenueSortRule) to instantiate.
        """
        self.txgroup: Final = txgroup
        self.networks: Final = (
            [networks] if isinstance(networks, int) else list(networks)
        )
        self.rule_type: Final = rule_type
        self._rule: Optional[TRule] = None

    @overload
    def __call__(
        self, txgroup_name: TxGroupName, networks: Optional[Networks] = None
    ) -> "SortRuleFactory":
        """Configure a nested sub‐group.

        Args:
            txgroup_name: Sub‐group name.
            networks: Optional network specification.
        """

    @overload
    def __call__(self, func: SortFunction) -> SortFunction:
        """Register a matching function.

        Args:
            func: The custom matching function.
        """

    def __call__(  # type: ignore [misc]
        self,
        func: Union[TxGroupName, SortFunction],
        networks: Optional[Networks] = None,
    ) -> Union["SortRuleFactory", SortFunction]:
        """Configure a nested sub‐group or register a matching function.

        Overloads:
            * If `func` is a string, returns a new factory for `txgroup:func`.
            * If `func` is callable, registers it as the match logic.

        Args:
            func: Sub‐group suffix (str) or a custom matching function.
            networks: Optional networks override (only valid when `func` is str).

        Raises:
            RuntimeError: If `networks` is passed when `func` is callable.
            ValueError: If `func` is neither str nor callable.

        See Also:
            :meth:`match`

        Examples:
            >>> fees = expense("Fees")
            >>> @fees("Gas")
            ... def match_gas(tx):
            ...     return tx.symbol == "ETH"
        """
        if isinstance(func, str):
            return SortRuleFactory(
                f"{self.txgroup}:{func}", networks or self.networks, self.rule_type
            )
        elif callable(func):
            if networks:
                raise RuntimeError("you can only pass networks if `func` is a string")
            if CHAINID in self.networks:
                self.__check_locked()
                self._rule = self.rule_type(txgroup=self.txgroup, func=func)
            return func
        raise ValueError(func)

    @property
    def rule(self) -> Optional[TRule]:
        """Return the created sort rule instance, if any.

        After decoration or a call to :meth:`match`, this property holds the
        concrete :class:`~dao_treasury.types.SortRule` instance.

        Examples:
            >>> @other_income("Interest")
            ... def match_i(tx):
            ...     return tx.value_usd > 100
        """
        return self._rule

    def match(
        self, func: None = None, **match_values: Any
    ) -> None:  # TODO: give this proper kwargs
        """Define static matching attributes for the sort rule.

        Call this method with keyword matchers corresponding to rule attributes
        (e.g., hash, from_address, symbol) to create a rule matching based on these values.

        Args:
            func: Must be None; a function match must use the decorator form.
            **match_values: Attribute values for matching (e.g., hash="0x123", symbol="DAI").

        Raises:
            ValueError: If `func` is not None.
            RuntimeError: If a matcher has already been set.

        See Also:
            :meth:`__call__`

        Examples:
            >>> ignore("Dust").match(symbol="WETH", from_address="0xAAA")
        """
        if func is not None:
            raise ValueError(
                f"You cannot pass a func here, call {self} with the function as the sole arg instead"
            )
        # Only instantiate when we're on an allowed network
        if CHAINID in self.networks:
            self.__check_locked()
            self._rule = self.rule_type(txgroup=self.txgroup, **match_values)
            self.locked = True

    def __check_locked(self) -> None:
        """Ensure that no matcher has already been registered.

        Raises:
            RuntimeError: If this factory already has a matcher assigned.
        """
        if self._rule is not None:
            raise RuntimeError(f"{self} already has a matcher")



from dao_treasury import TreasuryTx
from dao_treasury.sorting.factory import ignore
from dao_treasury.streams import llamapay


@ignore("LlamaPay")
def is_llamapay_stream_replenishment(tx: TreasuryTx) -> bool:
    if tx.to_address.address in llamapay.factories:  # type: ignore [operator]
        # We amortize these streams daily in the `llamapay` module, you'll sort each stream appropriately.
        return True

    # NOTE: not sure if we want this yet
    # Puling unused funds back from vesting escrow / llamapay
    # elif tx.from_address == "Contract: LlamaPay" and "StreamCancelled" in tx.events:
    #    if tx.amount > 0:
    #        tx.amount *= -1
    #    if tx.value_usd > 0:
    #        tx.value_usd *= -1
    #    return True
    return False



from typing import Final

import y.constants


CHAINID: Final = y.constants.CHAINID

ZERO_ADDRESS: Final = "0x0000000000000000000000000000000000000000"

# TODO: move disperse.app stuff from yearn-treasury to dao-treasury and then write a docs file
DISPERSE_APP: Final = (
    "0xD152f549545093347A162Dce210e7293f1452150",
    "0xd15fE25eD0Dba12fE05e7029C88b10C25e8880E3",
)
"""If your treasury sends funds to disperse.app, we create additional txs in the db so each individual send can be accounted for."""



from typing import TYPE_CHECKING, Awaitable, Callable, Iterable, Literal, NewType, Union

from y import Network

if TYPE_CHECKING:
    from dao_treasury.db import TreasuryTx
    from dao_treasury.sorting.rule import (
        CostOfRevenueSortRule,
        ExpenseSortRule,
        IgnoreSortRule,
        OtherExpenseSortRule,
        OtherIncomeSortRule,
        RevenueSortRule,
    )


Networks = Union[Network, Iterable[Network]]
"""Type alias for specifying one or more blockchain networks.

This can be a single :class:`~y.networks.Network` enum member or any iterable
of such members. It is used to restrict operations to specific networks when
configuring or querying treasury data.

Examples:
    Specify a single network:
        >>> net: Networks = Network.Mainnet

    Specify multiple networks:
        >>> nets: Networks = [Network.Mainnet, Network.Optimism]

See Also:
    :class:`~y.networks.Network`
"""


TopLevelCategory = Literal[
    "Revenue",
    "Cost of Revenue",
    "Expenses",
    "Other Income",
    "Other Expenses",
    "Ignore",
]
"""Literal type defining the allowed top‐level categories for treasury transactions.

Each transaction must be assigned to one of these categories in order to
generate financial reports.

Examples:
    Grouping a transaction as revenue:
        >>> cat: TopLevelCategory = "Revenue"

    Ignoring a transaction:
        >>> cat: TopLevelCategory = "Ignore"
"""


TxGroupDbid = NewType("TxGroupDbid", int)
"""NewType representing the primary key of a transaction group in the database.

This is the integer identifier returned by
:meth:`~dao_treasury.db.TxGroup.get_dbid`.

Examples:
    Get the database ID for the "Expenses" group:
        >>> dbid: TxGroupDbid = TxGroupDbid(3)
"""


TxGroupName = str
"""Alias for the human‐readable name of a transaction group.

Names are passed to
:meth:`~dao_treasury.db.TxGroup.get_dbid` 
to retrieve or create groups in the database.

Examples:
    Creating or retrieving the "Other Income" group:
        >>> name: TxGroupName = "Other Income"
        >>> dbid = TxGroup.get_dbid(name)
"""


SortFunction = Union[
    Callable[["TreasuryTx"], bool],
    Callable[["TreasuryTx"], Awaitable[bool]],
]
"""Type for a function or coroutine that determines if a transaction matches a rule.

A sorting function takes a single :class:`~dao_treasury.db.TreasuryTx`
instance and returns a boolean or an awaitable resolving to a boolean to
indicate whether the rule applies.

Examples:
    Synchronous matcher:
        >>> def is_large(tx: TreasuryTx) -> bool:
        ...     return tx.amount > 1000

    Asynchronous matcher:
        >>> async def has_tag(tx: TreasuryTx) -> bool:
        ...     return await some_external_check(tx.hash)

See Also:
    :class:`~dao_treasury.sorting.rule._SortRule.match`
"""


SortRule = Union[
    "RevenueSortRule",
    "CostOfRevenueSortRule",
    "ExpenseSortRule",
    "OtherIncomeSortRule",
    "OtherExpenseSortRule",
    "IgnoreSortRule",
]
"""Union of all built‐in sort rule classes.

Each rule assigns transactions to a specific category by matching on
transaction attributes or by invoking a custom function.

Examples:
    Define a list of available sort rules:
        >>> from dao_treasury.sorting.rule import SORT_RULES
        >>> rules: list[SortRule] = SORT_RULES

See Also:
    :class:`~dao_treasury.sorting.rule.RevenueSortRule`
    :class:`~dao_treasury.sorting.rule.ExpenseSortRule`
    :class:`~dao_treasury.sorting.rule.IgnoreSortRule`
    :class:`~dao_treasury.sorting.rule.CostOfRevenueSortRule`
    :class:`~dao_treasury.sorting.rule.OtherIncomeSortRule`
    :class:`~dao_treasury.sorting.rule.OtherExpenseSortRule`
"""
